@@ -208,7 +208,8 @@ class Flask(flask.Flask):
       method: `get`, `post`, `put`, `patch`, `delete`.
     - several additional methods are provided: `get_user_pass`,
       `user_in_group`, `check_password`, `hash_password`, `create_token`,
-      `get_user`, `current_user`, `clear_caches`, `register_cast`.
+      `get_user`, `current_user`, `clear_caches`, `register_cast`,
+      `register_object_perms`.
     """
 
     def __init__(self, *args, **kwargs):
@@ -239,6 +240,11 @@ class Flask(flask.Flask):
         """Add an object permission checker for a domain."""
         self._fsa.register_object_perms(d, f)
 
+    # decorator version
+    def object_perms(self, d):
+        """Decorator to add an object permission checker for a domain."""
+        return self._fsa.object_perms(d)
+
     # password management
     def check_password(self, pwd, ref):
         """Check whether password is ok wrt to current password manager."""
@@ -329,7 +335,7 @@ class FlaskSimpleAuth:
     #
     def _check_secure(self):
         if not request.is_secure and \
-           not (request.remote_addr.startswith("127.") or request.remote_addr == "::1"):
+           not (request.remote_addr.startswith("127.") or request.remote_addr == "::1"):  # pragma: no cover
             msg = f"insecure HTTP request on {request.remote_addr}, allow with FSA_SECURE=False"
             if self._secure:
                 log.error(msg)
@@ -450,12 +456,16 @@ class FlaskSimpleAuth:
 
     def get_user_pass(self, gup):
         """Set `get_user_pass` helper, can be used as a decorator."""
+        if self._get_user_pass:
+            log.warning("overriding already defined get_user_pass hook")
         self._get_user_pass = self._cache_function(gup, "g.")
         self._init_password_manager()
         return gup
 
     def user_in_group(self, uig):
         """Set `user_in_group` helper, can be used as a decorator."""
+        if self._user_in_group:
+            log.warning("overriding already defined user_in_group hook")
         self._user_in_group = self._cache_function(uig, "u.")
         return uig
 
@@ -465,8 +475,16 @@ class FlaskSimpleAuth:
 
     def register_object_perms(self, domain, checker: Callable):
         """Add an object permission helper for a domain."""
+        if domain in self._object_perms:
+            log.warning(f"overriding object permission checker for domain {domain}")
         self._object_perms[domain] = checker
-        # FIXME caching?
+
+    def object_perms(self, domain):
+        """Decorator to add an object permission hook for a domain."""
+        def annotate(fun):
+            self.register_object_perms(domain, fun)
+            return fun
+        return annotate
 
     def _check_object_perms(self, user, domain, oid, mode):
         """Tell whether user can access object oid in domain for mode."""
@@ -705,11 +723,9 @@ class FlaskSimpleAuth:
         self.blueprints = self._app.blueprints
         self.debug = False
         # portability tricks which generates mypy errors…
-        if hasattr(self._app, '_blueprint_order'):
-            # Flask 1.x
-            self._blueprint_order = self._app._blueprint_order  # type: ignore
-        elif hasattr(self._app, '_is_setup_finished'):
-            # Flask 2.0
+        if hasattr(self._app, '_blueprint_order'):  # Flask 1.x
+            self._blueprint_order = self._app._blueprint_order  # pragma: no cover
+        elif hasattr(self._app, '_is_setup_finished'):  # Flask 2.0
             self._is_setup_finished = self._app._is_setup_finished
             self.before_request_funcs = self._app.before_request_funcs
             self.after_request_funcs = self._app.after_request_funcs
@@ -718,7 +734,7 @@ class FlaskSimpleAuth:
             self.url_value_preprocessors = self._app.url_value_preprocessors
             self.template_context_processors = self._app.template_context_processors
         else:
-            log.warning("unexpected Flask version while dealing with blueprints?")
+            log.warning("unexpected Flask version while dealing with blueprints?")  # pragma: no cover
         #
         # caches
         #
@@ -1098,7 +1114,8 @@ class FlaskSimpleAuth:
                     break
             except FSAException as e:
                 lae = e
-            # FIXME other exceptions?
+            except Exception as e:  # pragma: no cover
+                log.error(f"internal error in {a} authentication: {e}")
 
         # even if not set, we say that the answer is the right one.
         self._user_set = True
@@ -1107,7 +1124,6 @@ class FlaskSimpleAuth:
         if required and not self._user:
             raise lae or FSAException("missing authentication", 401)
 
-        # log.debug(f"get_user({self._auth}): {self._user}")
         return self._user
 
     def current_user(self):
@@ -1186,7 +1202,7 @@ class FlaskSimpleAuth:
                     except FSAException as e:
                         return self._Resp(e.message, e.status)
 
-                if not self._user:
+                if not self._user:  # pragma no cover
                     return self._Resp("", 401)
 
                 return self._safe_call(path, "authenticate", fun, *args, **kwargs)
@@ -1303,12 +1319,13 @@ class FlaskSimpleAuth:
 
         return decorate
 
-    def _perm_auth(self, path, *perms):
+    def _perm_auth(self, path, first, *perms):
         """Decorator for per-object permissions."""
         # check perms wrt to recorded per-object checks
 
         # normalize tuples length to 3
-        perms = list(map(lambda a: (a + (None,)) if len(a) == 2 else a, perms))
+        perms = list(map(lambda a: (a + (first, None)) if len(a) == 1 else
+                                   (a + (None,)) if len(a) == 2 else a, perms))
 
         # perm checks
         for perm in perms:
@@ -1347,10 +1364,12 @@ class FlaskSimpleAuth:
                     except Exception as e:
                         log.error(f"internal error on {request.method} {request.path} permission {perm} check: {e}")
                         return self._Resp("internal error in permission check", self._server_error)
-                    if not isinstance(ok, bool):  # paranoid?
+                    if ok is None:
+                        return self._Resp(f"no such {domain} {val}", 404)
+                    elif not isinstance(ok, bool):  # paranoid?
                         log.error(f"type error on on {request.method} {request.path} permission {perm} check: {type(ok)}")
                         return self._Resp("internal error with permission check", self._server_error)
-                    if not ok:
+                    elif not ok:
                         return self._Resp("", 403)
 
                 # then call the initial function
@@ -1447,13 +1466,16 @@ class FlaskSimpleAuth:
 
         fun = view_func
 
-        # else only add needed filters on top of "fun"
+        # else only add needed filters on top of "fun", in reverse order
         need_authenticate = ALL in predefs or groups or perms
         need_parameters = len(fun.__code__.co_varnames) > 0
 
         if perms:
-            assert need_authenticate
-            fun = self._perm_auth(newpath, *perms)(fun)
+            if not need_parameters:
+                raise Exception("permissions require some parameters")
+            assert need_authenticate and need_parameters
+            first = fun.__code__.co_varnames[0]
+            fun = self._perm_auth(newpath, first, *perms)(fun)
         if need_parameters:
             fun = self._parameters(newpath)(fun)
         if groups:
@@ -1513,7 +1535,7 @@ class FlaskSimpleAuth:
         """Register a blueprint."""
 
         # lazy initialization
-        if not self._initialized:
+        if not self._initialized:  # pragma: no cover
             self.initialize()
 
         flask.Flask.register_blueprint(self, blueprint, **options)
