@@ -3,16 +3,18 @@
 #
 
 from anodb import DB  # type: ignore
-from FlaskSimpleAuth import Reference, Flask, Response
+from FlaskSimpleAuth import Reference, Flask, Response  # type: ignore
 
 # this reference will behave as a DB
 db = Reference()
 
+
 # always commit
 def db_commit(res: Response):
     db.commit()
     return res
 
+
 # module initialization
 def init_app(app: Flask):
     cf = app.config
