@@ -503,6 +503,7 @@ def test_route(client):
     check(400, client.get("/one/42"))   # missing "msg"
     check(404, client.get("/one/bad", data={"msg":"hi"}))  # bad "i" type
     check(403, client.get("/two", data={"LOGIN":"calvin"}))
+    check(518, client.get("/oops", data={"LOGIN":"calvin"}))
 
 def test_infer(client):
     res = check(200, client.get("/infer/1.000"))
@@ -953,10 +954,133 @@ def test_object_perms(client):
     check(403, client.get("/my/dad", data={"LOGIN": "calvin"}))
     check(200, client.get("/my/calvin", data={"LOGIN": "calvin"}))
     check(200, client.get("/my/hobbes", data={"LOGIN": "hobbes"}))
+    # no-such-user
+    check(404, client.get("/my/no-such-user", data={"LOGIN": "calvin"}))
 
-def test_object_perms_errors(client):
-    # TODO
-    assert True
+def test_object_perms_errors():
+    import AppFact as af
+    app = af.create_app(FSA_AUTH="fake")
+    @app.object_perms("known")
+    def is_okay(u: str, v: str, m: str):
+        log.debug(f"is_okay({u}, {v}, {m})")
+        if v == "fsa":
+            raise fsa.FSAException("oops-1", 518)
+        elif v == "ex":
+            raise Exception("oops-2")
+        elif v == "float":
+            return 3.15159
+        else:
+            return True
+    # triggers an overwrite warning
+    app.register_object_perms("known", is_okay)
+    # declaration time errors
+    try:
+        @app.get("/bad-perm-1", authorize=tuple())
+        def get_bad_perm_1(uid: int):
+            return "should not get there", 200
+        assert False, "should detect too short tuple"
+    except Exception as e:
+        assert "3 data" in str(e)
+    try:
+        @app.get("/bad-perm-2/<uid>", authorize=("unknown",))
+        def get_bad_perm_2_uid(uid: int):
+            return "should not get there", 200
+        assert False, "should detect unregistered permission domain"
+    except Exception as e:
+        assert "missing object permission" in str(e)
+    try:
+        @app.get("/bad-perm-3", authorize=("known", 3))
+        def get_bad_perm_3(uid: int):
+            return "should not get there", 200
+        assert False, "should detect bad variable name"
+    except Exception as e:
+        assert "unexpected identifier name type" in str(e)
+    try:
+        @app.get("/bad-perm-4/<uid>", authorize=("known", "uid", 3.14159))
+        def get_bad_perm_4_uid(uid: int):
+            return "should not get there", 200
+        assert False, "should detect bad mode type"
+    except Exception as e:
+        assert "unexpected mode type" in str(e)
+    try:
+        @app.get("/bad-perm-5", authorize=("known", "uid"))
+        def get_bad_perm_3(oid: int):
+            return "should not get there", 200
+        assert False, "should detect missing variable"
+    except Exception as e:
+        assert "missing function parameter uid" in str(e)
+    try:
+        @app.get("/bad-perm-6", authorize=("known", "uid"))
+        def get_bad_perm_3():
+            return "should not get there", 200
+        assert False, "should detect missing variable"
+    except Exception as e:
+        assert "permissions require some parameters" in str(e)
+    # run time errors
+    @app.get("/oops/<err>", authorize=("known", "err"))
+    def get_oops_err(err: str):
+        return "should not get there", 500
+    c = app.test_client()
+    res = c.get("/oops/fsa", data={"LOGIN": "calvin"})
+    assert res.status_code == 518 and b"oops-1" in res.data
+    res = c.get("/oops/ex", data={"LOGIN": "calvin"})
+    assert res.status_code == 500 and b"internal error in permission check" in res.data
+    res = c.get("/oops/float", data={"LOGIN": "calvin"})
+    assert res.status_code == 500 and b"internal error with permission check" in res.data
+
+def test_authorize_errors():
+    import AppFact as af
+    app = af.create_app()
+    try:
+        @app.get("/bad-authorize", authorize=[3.14159])
+        def get_bad_authorize():
+            return "should not get there", 200
+        assert False, "should detect bad authorize type"
+    except Exception as e:
+        assert "unexpected authorization" in str(e)
+    try:
+        @app.get("/bad-mix-1", authorize=["ANY", "ALL"])
+        def get_bad_mix_1():
+            return "should not get there", 200
+        assert False, "should detect ANY/ALL mix"
+    except Exception as e:
+        assert "ANY/ALL" in str(e)
+    try:
+        @app.get("/bad-mix-2", authorize=["ANY", "OTHER"])
+        def get_bad_mix_2():
+            return "should not get there", 200
+        assert False, "should detect ANY/other mix"
+    except Exception as e:
+        assert "other" in str(e)
+    try:
+        @app.get("/bad-mix-3", authorize=["ANY", ("foo", "id")])
+        def get_bad_mix_2():
+            return "should not get there", 200
+        assert False, "should detect ANY/other mix"
+    except Exception as e:
+        assert "object" in str(e)
+
+def test_group_errors():
+    import AppFact as af
+    def bad_uig(login, group):
+       if group == "ex":
+           raise fsa.FSAException("exception in user_in_group", 518)
+       elif group == "float":
+           return 3.14159
+       else:
+           return True
+    app = af.create_app(FSA_AUTH="fake", FSA_USER_IN_GROUP=bad_uig)
+    @app.get("/ex", authorize="ex")
+    def get_ex():
+        return "should not get here", 200
+    @app.get("/float", authorize="float")
+    def get_float():
+        return "should not get here", 200
+    c = app.test_client()
+    res = c.get("/ex", data={"LOGIN": "calvin"})
+    assert res.status_code == 518 and b"exception in user_in_group" in res.data
+    res = c.get("/float", data={"LOGIN": "calvin"})
+    assert res.status_code == 500 and b"internal error with user_in_group" in res.data
 
 # run some checks on AppFact, repeat to exercise caching
 def run_some_checks(c, n=10):
@@ -998,20 +1122,68 @@ def test_redis_cache():
         FSA_CACHE_OPTS={"host": "localhost", "port": 6379}).test_client() as c:
         run_some_checks(c)
 
-def test_lru_cache():
+def test_caches():
     import AppFact as af
-    with af.create_app(FSA_CACHE="lru").test_client() as c:
-        run_some_checks(c)
+    for cache in ["ttl", "lru", "lfu", "mru", "fifo", "rr", "dict"]:
+        log.debug(f"testing cache type {cache}")
+        with af.create_app(FSA_CACHE=cache).test_client() as c:
+            run_some_checks(c)
 
-def test_dict_cache():
-    import AppFact as af
-    with af.create_app(FSA_CACHE="dict").test_client() as c:
-        run_some_checks(c)
-
-def test_nosuch_cache():
+def test_no_such_cache():
     import AppFact as af
     try:
         af.create_app(FSA_CACHE="no-such-cache")
         assert False, "create app should fail"
     except Exception as e:
         assert "unexpected FSA_CACHE" in e.args[0]
+
+def test_warnings_and_errors():
+    def bad_gup_1(user: str):
+        raise fsa.FSAException("bad_gup_1", 518)
+    def bad_gup_2(user: str):
+        return 3.14159
+    import AppFact as af
+    app = af.create_app(
+        FSA_SUCH_DIRECTIVE="no-such-directive",
+        FSA_AUTH=["basic", "token"],
+        FSA_TOKEN_TYPE="fsa",
+        FSA_TOKEN_SIGN="signature-not-used-for-fsa-tokens",
+        FSA_FAKE_LOGIN="not-used-if-no-fake",
+        FSA_PARAM_USER="not-used-if-no-param",
+        FSA_PARAM_PASS="not-used-if-no-param",
+        FSA_PASSWORD_LEN=10,
+        FSA_PASSWORD_RE=[r"[0-9]"],
+        FSA_GET_USER_PASS=bad_gup_1,
+    )
+    app._fsa.initialize()
+    # password exceptions
+    try:
+        app.hash_password("short1")
+        assert False, "should be too short"
+    except fsa.FSAException as e:
+        assert e.status == 400 and "too short" in e.message
+    try:
+        app.hash_password("long-enough-but-missing-a-number")
+        assert False, "should not match re"
+    except fsa.FSAException as e:
+        assert e.status == 400 and "must match" in e.message
+    try:
+        app._fsa._check_password("calvin", "hobbes")
+        assert False, "should not get through"
+    except fsa.FSAException as e:
+        assert e.status == 518 and "bad" in e.message
+    # unused length warning
+    app = af.create_app(
+        FSA_TOKEN_TYPE="jwt",
+        FSA_TOKEN_LENGTH=8,  # no used if jwt
+        FSA_GET_USER_PASS=bad_gup_2,
+    )
+    app._fsa.initialize()
+    try:
+        app._fsa._check_password("calvin", "hobbes")
+        assert False, "should not get through"
+    except fsa.FSAException as e:
+        assert e.status == 500 and "internal error with get_user_pass" in e.message
+    # overwrite warning
+    fsa.register_cast("foo", lambda s: s)
+    fsa.register_cast("foo", lambda s: s)
